/*******************************************************************************
 * File name: IntelJ3.cpp
 * Description: Intel Flash burner
 * Author: Shiva	
 * Revision history:
 * 2010/10/01	Create this file.
 * 2012/05/25   Add comments - Kai
 * ****************************************************************************/

#include <sys/time.h>
#ifdef __MINGW32__
#include <winsock2.h>  // for Sleep()
#endif
#include "IntelJ3.h"
using namespace std;

#ifdef __MINGW32__
SOCKET sock;
#else
int sock;
#endif

/* In fast mode, IntelJ3 will skip some checks during burning */
int fast_mode = 0;
int debug_level = 0;
/* Test only */
int test_with_multiout = 0;
/* Measure performance - burning time */
bool g_measure_time = false;
/* Log file descriptor */
FILE *log = stdout;                  /* log file */

/* Reference: Numonyx Embedded Flash Memory (J3 v. D)
   Table 19: Status Register Bit Definitions:
   --------------------------------------------------------------------------------------------
   Bit   |Name                   | Description
   --------------------------------------------------------------------------------------------
   1     |Block-Locked Error     |0 = Block NOT locked during program or erase
         |                       |1 = Block locked during program or erase
   --------------------------------------------------------------------------------------------
   2     |Program Suspend Status |0 = Program suspend not in effect
         |                       |1 = Program suspend in effect
   --------------------------------------------------------------------------------------------
   3     |Error                  |0 = within acceptable limits during program or erase operation
         |                       |1 = not within acceptable limits during program or erase operation.
         |                       |    Operation aborted
   --------------------------------------------------------------------------------------------
   4:5   |Command Sequence Error |00 = Program or erase operation successful
         |                       |10 = Program error - operation aborted
         |                       |01 = Erase error - operation aborted
         |                       |11 = Command sequence error - command aborted
   --------------------------------------------------------------------------------------------
   6     |Erase Suspend Status   |0 = Erase suspend not in effect
         |                       |1 = Erase suspend in effect
   --------------------------------------------------------------------------------------------
   7     |Ready Status           |0 = Device is busy; SR[6:] are invalid (Not driven)
         |                       |1 = Device is ready SR[6:0] are valid
   --------------------------------------------------------------------------------------------
   Figure 22: Status Register Flowchart */
// TODO: modify this function to target-specific check status function
void check(unsigned data, const char *label)
{
  bool err_occur = false;

  if (data & 0x40)
  {
    fprintf(log, "\n%s - Erase suspend.\n", label);
    fflush(log);
    err_occur = true;
  }
  else if (data & 0x04)
  {
    fprintf(log, "\n%s - Program suspend.\n", label);
    fflush(log);
    err_occur = true;
  }
  else if (data & 0x30)
  {
    if ((data & 0x30) == 0x30)
    {
      fprintf(log, "\n%s - Command sequence error.\n", label);
      fflush(log);
    }
    else if ((data & 0x30) == 0x20)
    {
      fprintf(log, "\n%s - Erase error - operation aborted.\n", label);
      fflush(log);
    }
    else if ((data & 0x30) == 0x10)
    {
      fprintf(log, "\n%s - Program error - operation aborted.\n", label);
      fflush(log);
    }
    err_occur = true;
  }
  else if (data & 0x8)
  {
    fprintf(log, "\n%s - not within acceptable limits during program or erase operation.\n", label);
    fflush(log);
    err_occur = true;
  }
  else if (data & 2)
  {
    fprintf(log, "\n%s - Flash block locked.\n", label);
    fflush(log);
    err_occur = true;
  }

  if (err_occur == true)
  {
    terminate();
    exit(1);
  }
}

static void wait_done (unsigned base, const char *label)
{
  unsigned data = 0;
  unsigned time = 0;
  unsigned retry_count = 4096 * 2;  // NOTE: typical max block erase timeout = 4s

  while (data != 0x00800080) {
    outw(base,0x00700070);
    data=inw(base);

    time++;
    if (time > 16)
      SLEEP(1);

    if (time > retry_count)
      check(data, label);
  }
}

/* Reference: Numonyx Embedded Flash Memory (J3 v. D) */
/* Table 30: Command Bus Operation */
/* Command                |Address Bus     |Data Bus   */
/* --------------------------------------------------- */
/* Read Identifier Codes  |Device Address  |0x90       */
/* --------------------------------------------------- */
// TODO: modify this function to target-specific function
void Flash_ReadID(unsigned flash_type, unsigned base, unsigned BusWidth, unsigned* DDI1, unsigned* DDI2)
{
    if (flash_type == Intel_Type)
    {
        switch(BusWidth)
        {
        case FLASH_SINGLE: // 8 bit
            outw(base,0x90);
            *DDI1 = inw(base+0x00);
            *DDI2 = inw(base+0x01);
            outw(base,0xff);  // Issue "Read Array" to exit Read Status Register mode
            break;
        case FLASH_DOUBLE: // double 16 bit
            outw(base,0x9090);
            *DDI1 = inw(base+0x00);
            *DDI2 = inw(base+0x02);
            outw(base,0xffff); // Issue "Read Array" to exit Read Status Register mode
            break;
        case FLASH_FOUR:// four 32 bit
            outw(base,0x00500050);
            outw(base,0x00900090);
            *DDI1 = inw(base+0x00);
            *DDI2 = inw(base+0x04);
            outw(base,0x00ff00ff); // Issue "Read Array" to exit Read Status Register mode
            break;
        }
    }
}

/* Reference: Numonyx Embedded Flash Memory (J3 v. D) */
/* Table 30: Command Bus Operation */
/* Command                |Address Bus     |Data Bus   |Address Bus     |Data Bus   */
/* -------------------------------------------------------------------------------- */
/* Clear Status Register  |Device Address  |0x50       |---             |---        */
/* Unlock Block           |Device Address  |0x60       |Device Address  |0xD0       */
/* Read Status Register   |Device Address  |0x70       |---             |---        */
/* -------------------------------------------------------------------------------- */
// TODO: modify this function to target-specific function
void Flash_UnLock(unsigned flash_type, unsigned base, unsigned BusWidth)
{
    if (flash_type == Intel_Type)
    {
      outw(base,0x00600060);
      outw(base,0x00d000d0);

      wait_done(base, "Flash_UnLock");
      outw(base,0x00ff00ff);
    }
}


/* Reference: Numonyx Embedded Flash Memory (J3 v. D) */
/* Table 30: Command Bus Operation */
/* Command                |Address Bus     |Data Bus   |Address Bus     |Data Bus   */
/* -------------------------------------------------------------------------------- */
/* Lock Block             |Device Address  |0x60       |Device Address  |0x01       */
/* Read Status Register   |Device Address  |0x70       |---             |---        */
/* -------------------------------------------------------------------------------- */
// TODO: modify this function to target-specific function
void Flash_Lock(unsigned flash_type, unsigned base, unsigned BusWidth)
{
    if (flash_type == Intel_Type)
    {
      outw(base,0x00600060);
      outw(base,0x00010001);

      // read status register
      wait_done (base, "Flash_Lock");
    }
}

/* Reference: Numonyx Embedded Flash Memory (J3 v. D) */
/* Table 30: Command Bus Operation */
/* Command                |Address Bus     |Data Bus   |Address Bus     |Data Bus   */
/* -------------------------------------------------------------------------------- */
/* Block Erase            |Block Address   |0x20       |Block Address   |0xD0       */
/* Read Array             |Device Address  |0xFF       |---             |---        */
/* Read Status Register   |Device Address  |0x70       |---             |---        */
/* -------------------------------------------------------------------------------- */
// TODO: modify this function to target-specific function
void Flash_ChipErase(unsigned flash_type, unsigned base, unsigned BusWidth)
{
    if (flash_type == Intel_Type)
    {
      outw(base,0x00200020);
      outw(base,0x00d000d0);
      SLEEP (32);  // typical erase time = (2^10)ms = 1s

      wait_done (base, "Flash_ChipErase");
    }
}

/* Reference: Numonyx Embedded Flash Memory (J3 v. D) */
/* Table 30: Command Bus Operation */
/* Command                |Address Bus     |Data Bus   |Address Bus     |Data Bus   */
/* -------------------------------------------------------------------------------- */
/* Word Program           |Device Address  |0x40       |Device Address  |Array Data */
/* Read Status Register   |Device Address  |0x70       |---             |---        */
/* Read Array             |Device Address  |0xFF       |---             |---        */
/* -------------------------------------------------------------------------------- */
// TODO: modify this function to target-specific function - write single word
int Flash_ProgramWord(unsigned flash_type, unsigned base, volatile unsigned address, volatile unsigned data)
{
#ifndef FPGA 
    if (flash_type == Intel_Type)
#endif // #ifndef FPGA
    {
       if (outw(base, 0x00400040) < 0)
         return -1;

       outw(address, data);

       wait_done (base, "Flash_ProgramWord");
    }

    return 0;
}

/* Reference: Numonyx Embedded Flash Memory (J3 v. D) */
/* Table 30: Command Bus Operation */
/* Command                |Address Bus     |Data Bus   |Address Bus     |Data Bus   */
/* -------------------------------------------------------------------------------- */
/* Buffered Program       |Word Address    |0xE8       |Device Address  |0xD0       */
/* Read Status Register   |Device Address  |0x70       |---             |---        */
/* Read Array             |Device Address  |0xFF       |---             |---        */
/* -------------------------------------------------------------------------------- */
/* Figure 21: Write to Buffer Flowchart */
/* NOTE: Max is 16 words in one time for this flash */
// TODO: modify this function to target-specific function - write multiple words
int Flash_ProgramMultiWord(unsigned flash_type, unsigned base, volatile unsigned address, unsigned char *buffer, unsigned size)
{
    volatile unsigned int s_reg = 0;
    volatile unsigned int block_addr = (address/BLK_SIZE)*BLK_SIZE;
    volatile unsigned int bytes = size % 4;
    volatile unsigned int word;
    volatile unsigned int time = 0;
 
    /* not word align */
    if (size % 4 != 0) size = size - bytes;

    if (flash_type == Intel_Type)
    {
       if (size != 0)
       {
         time = 0;
         do {
           outw(block_addr, 0x00e800e8);
           s_reg = inw(base);

           time++;
           if (time >= 1000)
             check(s_reg, "Flash_ProgramMultiWord");

         } while (s_reg != 0x00800080);  // check SR7

         word = (size/4) - 1;
         word = (word << 16) + word;
         outw(block_addr, word);
#if 0
         for(unsigned int i = 0; i < (size/4); i++) 
         {
           data = ((((unsigned int)buffer[4*i+3]<< 24) & 0xFF000000) |
                   (((unsigned int)buffer[4*i+2]<< 16) & 0x00FF0000) |
                   (((unsigned int)buffer[4*i+1]<< 8) & 0x0000FF00) |
                   (((unsigned int)buffer[4*i]<< 0) & 0x000000FF));
           outw(address + 4*i, data);
         }
#else
         if (fastout(address, size, (char*)buffer) < 0)
           return -1;
#endif
         outw(block_addr, 0x00d000d0);
         s_reg = 0;

         if (fast_mode == 0)
           wait_done (base, "Flash_ProgramMultiWord_Finish");
       }

       if (bytes)
       {
         word = 0;
         for (unsigned int j = 0; j <= bytes; j++) 
           word = word | (buffer[size+j] << (j*8));

         if (Flash_ProgramWord (flash_type, base, address+size, word) < 0)
           return -1;
       }
    }
    else
    {
      printf("not intel flash !!\n");
      return -1;
    }

    return 0;
}

/* main function to burn image to flash */
/* Use Flash_ProgramMultiWord () to write 64 bytes one time */
// TODO: modify this function to target-specific function
int Burn_image(unsigned flash_type, unsigned int address, unsigned char* start, unsigned int size, unsigned int base)
{
  unsigned int end = address + size;
  unsigned int buf_size;
  int result = 0;

  while (address < end)
  {
    if (flash_type == Intel_Type)
    {
      // write 64 bytes one time
      buf_size = ((end - address) > 64) ? 64 : (end - address);
      if (Flash_ProgramMultiWord(flash_type, base, address, start, buf_size) < 0)
      {
        result = -1;
        break;
      }

      start += buf_size;
      address += buf_size;
    }

    if (address % 4096 == 0)
    {
      fprintf(log, ".");
      fflush(log);
    }
  }
  outw(base, 0x00ff00ff);

  return result;
}

/* Help menu */
unsigned int usage(void)
{
  printf( "--image:\t\tImage name to burn\n");
  printf( "--target:\t\tBoard type to find ROM start address (xc5/ag101/ag101p_16mb)\n");
  printf( "--host:\t\t\tHost name/ip to connect with ICEman\n");
  printf( "--port:\t\t\tPort number to connect with ICEman (default to 2354)\n");
  printf( "--base:\t\t\tFlash base address (default to 0x0)\n");
  printf( "--addr:\t\t\tFlash target address to write (default to 0x0)\n");
  printf( "--preserve:\t\tPreserve the content in first and last erase blocks\n");
  printf( "--log:\t\t\tThe log file to store output message (default to stdout)\n");
  printf( "--reset-and-run:\tReset target\n");
  printf( "--reset-and-hold:\tReset target and stop at $IVB\n");
  printf( "--verify:\t\tVerify after flash burning\n");
  printf( "--verify-only:\t\tUse image file to verify content of ROM, no burn\n");
  printf( "--version:\t\tShow flash burning version\n");
  printf( "--unlock:\t\tUnlock flash before burning\n");
  printf( "--lock:\t\t\tLock flash after burning\n");
  printf( "--fast:\t\t\tBurning flash in fast mode\n");
  printf( "--erase-all:\t\tErase entire flash before burning\n");
  printf( "--help:\t\t\tThe usage for IntelJ3 Burner\n");
  printf( "--measure-time:\t\tEstimate total time to burn image to ROM\n");
  return 0;
}

/* Performance utility functions - to calculate the difference of a_timeval_begin and a_timeval_end */
static void timeval_diff (struct timeval *a_result, struct timeval *a_timeval_begin, struct timeval *a_timeval_end)
{
  if (a_timeval_end->tv_usec >= a_timeval_begin->tv_usec)
  {
    a_result->tv_usec = (a_timeval_end->tv_usec - a_timeval_begin->tv_usec);
    a_result->tv_sec = (a_timeval_end->tv_sec - a_timeval_begin->tv_sec);
    if (a_result->tv_usec >= 1000000)
    {
      a_result->tv_usec -= 1000000;
      a_result->tv_sec += 1;
    }
  }
  else
  {
    a_result->tv_usec = (a_timeval_end->tv_usec + (1000000 - a_timeval_begin->tv_usec));
    a_result->tv_sec = (a_timeval_end->tv_sec - a_timeval_begin->tv_sec - 1);
  }
}

/* main function 
 *  1. parse parameters
 *  2. initial socket
 *  3. reset AICE 
 *  4. issue RESET_HOLD to let target board restore to reset status 
 *     avoid boot code to damage flash registers' settings 
 *  5. call target-dependent initial function and calculate flash base address (Device Address) 
 *     xc5()/p24()/ag101()/ag301p() 
 *  6. read image binary data from --image specified file 
 *  7. check flash ID 
 *  8. if user specifies --verify-only, verify the target with the image data and exit 
 *  9. unlock flash
 * 10. preserve first/last blocks if needed (As start/end addresses are not block-aligned)
 * 11. erase flash
 * 12. burn image - call Burn_image()
 * 13. verify
 * 14. lock flash */
int main(int argc, char **argv)
{
  int c = 0;
  char host[64] = "127.0.0.1";        /* default connecting host */
  unsigned int port = 2354;         /* default connecting port */
  unsigned int base = 0x80400000;    /* flash base address */
  unsigned int address = 0x80400000; /* default write to flash base */
  unsigned int smc_base = 0x80400000;
  unsigned int size;                 /* image size */
  unsigned int erase_size;           /* erase size */
  unsigned int end;                  /* binary end address */
  unsigned int pt, i, j;
  unsigned int id1, id2;             /* flash id */ 
  unsigned int first_size, last_size;             
  unsigned char* flash = NULL;       /* image pointer */
  unsigned char* start = NULL;       /* image pointer */
  unsigned char first[BLK_SIZE], last[BLK_SIZE];
  FILE *image = NULL;                /* image to burn */
  char flash_type = Intel_Type;
  char cmd = 0;
  char *name = new char[1024];
  int preserve = 0;
  int specify_base = 0;
  int specify_address = 0;
  int xc5_enable = 0;
  int ag301p_enable = 0;
  int ag101_enable = 0;
  int p24_enable = 0;
  int verify = 0;
  int lock = 0;
  int unlock = 0;
  int erase_all = 0;
  int verify_only = 0;
  struct timeval erase_begin_time;
  struct timeval erase_end_time;
  struct timeval erase_total_time;
  double erase_time_sec;
  struct timeval burn_begin_time;
  struct timeval burn_end_time;
  struct timeval burn_total_time;
  double burn_time_sec;
  unsigned int total_burn_size;
  struct timeval verify_begin_time;
  struct timeval verify_end_time;
  struct timeval verify_total_time;
  double verify_time_sec;

  /* parameter handler */
  while(1)
  {
    int option_index;
    c = getopt_long(argc, argv, "ugvcVtoehzyTm:l:x:k:H:p:a:b:f:i", long_option, &option_index);
    if (c == EOF) break;

    switch (c)
    {
      case 'u':
	erase_all = 1;
	break;
      case 'z':
	lock = 1;
	break;
      case 'g':
	fast_mode = 1;
	break;
      case 'y':
	unlock = 1;
	break;
      case 'v':
	verify = 1;
	break;
      case 'c':
        verify_only = 1;
        break;
      case 'l':
	if ((log = fopen(optarg, "w"))== NULL) {
          fprintf (log, "Error: can't open file %s\n", optarg);
          fflush(log);
          exit(1);
        }
        break;
      case 's':
        fprintf (log, "Branch%s\n", BRANCH_NAME);
        fprintf (log, "%s\n", COMMIT_ID);
        fflush(log);
        exit(0);
      case 'V':
        fprintf (log, "IntelJ3 Burner BUILD_ID: %d\n", BUILD_ID);
        fflush(log);
        exit(0);
	break;
      case 'x':
        if (strncmp(optarg, "xc5", 6) == 0)
        {
          xc5_enable = 1;
        } else if (strncmp(optarg, "ag101", 6) == 0)
        {
  	  ag101_enable = 1;
        } else if (strncmp(optarg, "ag301p", 6) == 0)
        {
  	  ag301p_enable = 1;
        } else if (strncmp(optarg, "ag101p_16mb", 6) == 0)
        {
  	  p24_enable = 1;
        }
        // TODO: add target type
        break;
      case 't':
        cmd = RESET_TARGET;
        break;
      case 'o':
        cmd = RESET_HOLD;
        break;
      case 'e':
        preserve = 1;
        break;
      case 'H':
	strcpy(host, optarg);
	break;
      case 'p':
        port = strtol(optarg, NULL, 0);
        break;
      case 'a':
        sscanf(optarg,"0x%x", &address);
        specify_address = 1;
        break;
      case 'b':
        sscanf(optarg,"0x%x", &base);
        specify_base = 1;
        break;
      case 'f':
        if (strncmp(optarg, "mx", 6) == 0)
        {
          flash_type = MX_Type;
        } else if (strncmp(optarg, "sst", 6) == 0)
        {
          flash_type = SST_Type;
        } else if (strncmp(optarg, "intel", 6) == 0)
        {
          flash_type = Intel_Type;
        } else if (strncmp(optarg, "amd", 6) == 0)
        {
          flash_type = AMD_Type;
        }
        break;
      case 'i':
        if (strncmp(optarg, "/cygdrive/", 10) == 0)
        {
          name[0] = optarg[10];
          name[1] = ':';
          name[2] = '\0';
          strcat(name, optarg + 11);
        }
        else name = optarg;
	if ((image = fopen(name, "rb"))== NULL) {
                fprintf (log, "Error: can't open file %s\n", optarg);
                fflush(log);
                exit(1);
        }
        break;
      case 'm':
        g_measure_time = true;
        break;
      case 'T':
        test_with_multiout = 1;
        break;
      case 'h':
      case '?':
      default:
        usage();
        exit(0);
        break;
    }
  }

  fprintf (log, "IntelJ3 Burner BUILD_ID: %d\n", BUILD_ID);
  fflush(log);

  initial_socket(host, port);
#ifdef __MINGW32__
  Sleep (200);
#else
  usleep (200000);
#endif

  // init AICE
  if (send_cmd(RESET_AICE) < 0)
  {
    close_connection ();
    exit(0);
  }

  if (cmd != 0)
  {
    send_cmd (cmd);
    close_connection ();
    exit(0);
  }

  // reset-and-hold to init target
  send_cmd (RESET_HOLD);
  smc_base = 0;

  // TODO: write target-specific initial function
  if (xc5_enable)
  {
    smc_base = xc5();
  }
  else if (ag101_enable)
  {
    smc_base = ag101();
  }
  else if (ag301p_enable)
  {
    smc_base = ag301p();
  }
  else if (p24_enable)
  {
    smc_base = p24();
  }
  else
  {
    fprintf (log, "WARNING: No target specified.  ");
    if (specify_base == 0)
      fprintf (log, "Use 0x80400000 as base address.\n");
    else
      fprintf (log, "Use 0x%8x as base address.\n", base);
  }

  if (specify_address == 0) address = smc_base;
  if (specify_base == 0) base = smc_base;

  /* get burn image content */
  flash = get_image (image, &size);
  fclose (image);
  start = flash;

  /* handle signal */
  signal(SIGTERM, handle_int);
  signal(SIGINT, handle_int);
#ifndef __MINGW32__
  signal(SIGKILL, handle_int);
#endif
  /* calculate address */
  end = address + size;
  pt = (address/BLK_SIZE) * BLK_SIZE;
  erase_size = ((size/BLK_SIZE) + 1) * BLK_SIZE;

  fprintf(log, "Flash base address: 0x%x\n", base);

  /* check device ID */
  Flash_ReadID(flash_type, base, FLASH_FOUR, &id1, &id2);
  // TODO: change 0x890089 & 0x180018 to target-specific flash ID
  if (id1 == 0x890089 && id2 == 0x180018) {
    fprintf(log, "burn data to intel flash\n");
    fflush(log);
  }
  else
  {
    terminate();
    fprintf(log, "Not intel flash !!\n");
    fprintf(log, "ID1: 0x%x, ID2: 0x%x\n", id1, id2);
    fflush(log);
    return 0;
  }
  /* verify only */
  if (verify_only)
  {
    fprintf(log, "\nverifying");
    fflush(log);
    if (Verification(address, start, size))
    {
      fprintf(log, "Verify fail !!!\n");
      fflush(log);
    }
    else 
    {
      fprintf(log, "\nVerify success.\n");
      fflush(log);
    }

    terminate ();
    return 0;
  }

  /* unlock flash */
  if (unlock)
  {
    fprintf(log, "unlocking");
    fflush(log);
    Flash_UnLock(flash_type, pt, FLASH_FOUR);
    fprintf(log, "\n");
    fflush(log);
  }
  /* record the content in first and last erase block */
  if (preserve)
  {
    first_size = address - pt;
    last_size = erase_size - size - first_size;

    if (first_size)
    {
      Record_content(pt, first, first_size);
    }
    if (last_size)
    {
      Record_content(end, last, last_size);
    }
  }

  /* block erase */
  if (erase_all)
  {
    pt = base;
    erase_size = FLASH_SIZE;
    if (p24_enable) erase_size = 0x100000;
  }
  fprintf(log, "erase from address = %x\n", pt);
  fflush(log);

  if (g_measure_time == true)
    gettimeofday (&erase_begin_time, NULL);
  for (i = 0, j = pt; i < erase_size/BLK_SIZE; i++)
  {
    fprintf(log, "erasing block %d (0x%x ~ 0x%x)\n", i+1, j, j + BLK_SIZE);
    fflush(log);
    Flash_ChipErase(flash_type, j, FLASH_FOUR);
    j += BLK_SIZE;
  }
  outw(base,0x00ff00ff);

  if (g_measure_time == true)
  {
    gettimeofday (&erase_end_time, NULL);

    timeval_diff (&erase_total_time, &erase_begin_time, &erase_end_time);
    erase_time_sec = erase_total_time.tv_sec + (double)erase_total_time.tv_usec / 1000000;

    fprintf(log, "\nTotal erase time: %lu sec, %lu usec\n", 
        erase_total_time.tv_sec, erase_total_time.tv_usec);
    fprintf(log, "Average erase rate: %.2lf KBytes/s\n", (double)(i * BLK_SIZE >> 10) / erase_time_sec);
    fflush(log);
  }

  if (debug_level >= 3)
  {
    fprintf(log, "address = %x, base = %x, flash type = %d, size = %x\n", address, base, flash_type, size); 
    fflush(log);
  }
  fprintf(log, "burn flash from 0x%x to 0x%x\n", address ,end);
  fflush(log);
  fprintf(log, "burning");
  fflush(log);

  if (g_measure_time == true)
    gettimeofday (&burn_begin_time, NULL);
  /* burn image */
  if (Burn_image(flash_type, address, flash, size, base) < 0)
  {
    close_connection ();
    return 0;
  }

  total_burn_size = size;

  /* burn first and last block content back to flash */
  if (preserve)
  {
    if (first_size)
    {
      if (Burn_image(flash_type, pt, first, first_size, base) < 0)
      {
        close_connection ();
        return 0;
      }
      total_burn_size += first_size;
    }
    if (last_size)
    {
      if (Burn_image(flash_type, end, last, last_size, base) < 0)
      {
        close_connection ();
        return 0;
      }

      total_burn_size += last_size;
    }
  }
  if (g_measure_time == true)
  {
    gettimeofday (&burn_end_time, NULL);

    timeval_diff (&burn_total_time, &burn_begin_time, &burn_end_time);
    burn_time_sec = burn_total_time.tv_sec + (double)burn_total_time.tv_usec / 1000000;

    fprintf(log, "\n\nTotal burn time: %lu sec, %lu usec\n", 
        burn_total_time.tv_sec, burn_total_time.tv_usec);
    fprintf(log, "Average burn rate: %.2lf KBytes/s\n", (double)(total_burn_size >> 10) / burn_time_sec);
    fflush(log);
  }

  /* verify */
  if (verify)
  {
    if (g_measure_time == true)
      gettimeofday (&verify_begin_time, NULL);

    fprintf(log, "\nverifying");
    fflush(log);
    if (Verification(address, start, size))
    {
      fprintf(log, "Verify fail !!!\n");
      fprintf(log, "Please reburn again.\n");
      fflush(log);
    }
    else 
    {
      fprintf(log, "\nVerify success.\n");
      fprintf(log, "Flash burning done.\n");
      fflush(log);
    }

    if (g_measure_time == true)
    {
      gettimeofday (&verify_end_time, NULL);

      timeval_diff (&verify_total_time, &verify_begin_time, &verify_end_time);
      verify_time_sec = verify_total_time.tv_sec + (double)verify_total_time.tv_usec / 1000000;

      fprintf(log, "\n\nTotal verify time: %lu sec, %lu usec\n",
          verify_total_time.tv_sec, verify_total_time.tv_usec);
      fprintf(log, "Average verify rate: %.2lf KBytes/s\n", (double)(size >> 10) / verify_time_sec);
      fflush(log);
    }
  }
  else
  {
    fprintf(log, "\nFlash burning done.\n");
    fflush(log);
  } 

  /* lock flash */
  if (lock)
  {
    fprintf(log, "locking");
    fflush(log);
    for (i = 0, j = pt; i < erase_size/BLK_SIZE; i++)
    {
      fprintf(log, ".");
      fflush(log);
      Flash_Lock(flash_type, j, FLASH_FOUR);
      j += BLK_SIZE;
    }
    outw(base,0x00ff00ff);
    fprintf(log, "\n");
    fflush(log);
  }

  /* send terminate signal */
  terminate();
  return 0;
}

